
from logging import Logger
from logging import getLogger

from untangle import Element

from pyutmodel.PyutActor import PyutActor
from pyutmodel.PyutUseCase import PyutUseCase

from ogl.OglActor import OglActor
from ogl.OglUseCase import OglUseCase

from untanglepyut.Common import GraphicInformation
from untanglepyut.Common import toGraphicInfo
from untanglepyut.Common import createUntangledOglActors
from untanglepyut.Common import createUntangledOglUseCases

from untanglepyut.Types import UntangledOglActors
from untanglepyut.Types import UntangledOglUseCases

from untanglepyut.BaseUnTangle import BaseUnTangle
from untanglepyut.UnTanglePyut import UnTanglePyut


class UnTangleUseCaseDiagram(BaseUnTangle):
    """
        <PyutDocument type="USECASE_DIAGRAM" title="Use-Cases" scrollPositionX="0" scrollPositionY="0" pixelsPerUnitX="20" pixelsPerUnitY="20">
            <GraphicActor width="87" height="114" x="293" y="236">
                <Actor id="1" name="BasicActor" filename=""/>
            </GraphicActor>
            <GraphicUseCase width="100" height="60" x="575" y="250">
                <UseCase id="2" name="Basic Use Case" filename=""/>
            </GraphicUseCase>
            <GraphicLink srcX="379" srcY="286" dstX="575" dstY="280" spline="False">
                <LabelCenter x="555" y="281"/>
                <LabelSrc x="555" y="281"/>
                <LabelDst x="555" y="281"/>
                <Link name="Kicks Butt" type="ASSOCIATION" cardSrc="" cardDestination="" bidir="False" sourceId="1" destId="2"/>
            </GraphicLink>
        </PyutDocument>
    """

    def __init__(self):

        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._untangledOglActors:   UntangledOglActors   = createUntangledOglActors()
        self._untangledOglUseCases: UntangledOglUseCases = createUntangledOglUseCases()
        self._untanglePyut:         UnTanglePyut         = UnTanglePyut()

    def unTangle(self, pyutDocument: Element):
        """

        Args:
            pyutDocument:
        """

        self._untangledOglActors   = self._unTangleOglActors(pyutDocument=pyutDocument)
        self._untangledOglUseCases = self._unTangleOglUseCases(pyutDocument=pyutDocument)

    @property
    def oglActors(self) -> UntangledOglActors:
        return self._untangledOglActors

    @property
    def oglUseCases(self) -> UntangledOglUseCases:
        return self._untangledOglUseCases

    def _unTangleOglActors(self, pyutDocument: Element) -> UntangledOglActors:
        untangledOglActors: UntangledOglActors = createUntangledOglActors()

        graphicActors: Element = pyutDocument.get_elements('GraphicActor')
        for graphicActor in graphicActors:
            graphicInfo: GraphicInformation = toGraphicInfo(graphicActor)
            oglActor:    OglActor           = OglActor(w=graphicInfo.width, h=graphicInfo.height)
            oglActor.SetPosition(x=graphicInfo.x, y=graphicInfo.y)

            self._updateModel(oglObject=oglActor, graphicInformation=graphicInfo)

            pyutActor: PyutActor = self._untanglePyut.actorToPyutActor(graphicActor=graphicActor)

            oglActor.pyutObject = pyutActor

            untangledOglActors.append(oglActor)

        return untangledOglActors

    def _unTangleOglUseCases(self, pyutDocument: Element) -> UntangledOglUseCases:

        untangledOglUseCases: UntangledOglUseCases = createUntangledOglUseCases()

        graphicUseCases: Element = pyutDocument.get_elements('GraphicUseCase')
        for graphicUseCase in graphicUseCases:
            graphicInfo: GraphicInformation = toGraphicInfo(graphicUseCase)
            oglUseCase:  OglUseCase         = OglUseCase(w=graphicInfo.width, h=graphicInfo.height)

            oglUseCase.SetPosition(x=graphicInfo.x, y=graphicInfo.y)

            self._updateModel(oglObject=oglUseCase, graphicInformation=graphicInfo)

            pyutUseCase: PyutUseCase = self._untanglePyut.useCaseToPyutUseCase(graphicUseCase=graphicUseCase)

            oglUseCase.pyutObject = pyutUseCase

            untangledOglUseCases.append(oglUseCase)

        return untangledOglUseCases



from dataclasses import dataclass
from typing import List
from typing import NewType
from typing import cast

from logging import Logger
from logging import getLogger

from untangle import Element

from pyutmodel.PyutField import PyutField
from pyutmodel.PyutObject import PyutObject
from pyutmodel.PyutUseCase import PyutUseCase
from pyutmodel.PyutActor import PyutActor
from pyutmodel.PyutInterface import PyutInterface
from pyutmodel.PyutLink import PyutLink
from pyutmodel.PyutLinkType import PyutLinkType
from pyutmodel.PyutMethod import PyutParameters
from pyutmodel.PyutMethod import SourceCode
from pyutmodel.PyutParameter import PyutParameter
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutMethod import PyutModifiers
from pyutmodel.PyutModifier import PyutModifier
from pyutmodel.PyutType import PyutType
from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutNote import PyutNote
from pyutmodel.PyutText import PyutText
from pyutmodel.PyutStereotype import PyutStereotype
from pyutmodel.PyutVisibilityEnum import PyutVisibilityEnum
from pyutmodel.PyutDisplayParameters import PyutDisplayParameters
from pyutmodel.PyutSDInstance import PyutSDInstance
from pyutmodel.PyutSDMessage import PyutSDMessage

from untanglepyut.Common import secureInteger
from untanglepyut.Common import str2bool

Elements = NewType('Elements',  List[Element])

UntangledPyutMethods = NewType('UntangledPyutMethods', List[PyutMethod])
UntangledPyutFields  = NewType('UntangledPyutFields',  List[PyutField])


@dataclass
class ConvolutedPyutSDMessageInformation:
    """
    This class is necessary because I do not want to mix Ogl and PyutModel code;  Unfortunately,
    the IDs of the PyutSDInstance are buried and require a lookup

    """
    pyutSDMessage: PyutSDMessage = cast(PyutSDMessage, None)
    sourceId:      int           = -1
    destinationId: int           = -1


class UnTanglePyut:
    """
    Converts PyutModel XML to Pyut Objects
    """
    def __init__(self):

        self.logger: Logger = getLogger(__name__)

    def classToPyutClass(self, graphicClass: Element) -> PyutClass:
        classElement: Element = graphicClass.Class

        pyutClass: PyutClass = PyutClass()

        pyutClass = cast(PyutClass, self._addPyutObjectAttributes(pyutElement=classElement, pyutObject=pyutClass))

        displayParameters: PyutDisplayParameters = PyutDisplayParameters.toEnum(classElement['displayParameters'])

        showStereotype:    bool = bool(classElement['showStereotype'])
        showFields:        bool = bool(classElement['showFields'])
        showMethods:       bool = bool(classElement['showMethods'])
        stereotypeStr:     str  = classElement['stereotype']

        pyutClass.displayParameters = displayParameters

        pyutClass.displayStereoType = showStereotype
        pyutClass.showFields     = showFields
        pyutClass.showMethods    = showMethods

        pyutClass.description = classElement['description']
        pyutClass.stereotype = PyutStereotype(name=stereotypeStr)

        pyutClass.methods = self._methodToPyutMethods(classElement=classElement)
        pyutClass.fields  = self._fieldToPyutFields(classElement=classElement)

        return pyutClass

    def textToPyutText(self, graphicText: Element) -> PyutText:
        """
        Parses Text elements
        Args:
            graphicText:   Of the form:   <Text id="3" content="I am standalone text"/>

        Returns: A PyutText Object
        """
        textElement: Element  = graphicText.Text
        pyutText:    PyutText = PyutText()

        pyutText.id     = textElement['id']
        pyutText.content = textElement['content']

        return pyutText

    def noteToPyutNote(self, graphicNote: Element) -> PyutNote:
        """
        Parse Note elements
        Args:
            graphicNote: of the form:  <Note id="2" content="I am a UML Note" filename=""/>

        Returns: A PyutNote Object
        """
        noteElement: Element = graphicNote.Note
        pyutNote: PyutNote = PyutNote()

        pyutNote = cast(PyutNote, self._addPyutObjectAttributes(pyutElement=noteElement, pyutObject=pyutNote))

        pyutNote.content = noteElement['content']
        return pyutNote

    def interfaceToPyutInterface(self, interface: Element) -> PyutInterface:

        interfaceId: int = int(interface['id'])
        name:        str = interface['name']
        description: str = interface['description']

        pyutInterface: PyutInterface = PyutInterface(name=name)
        pyutInterface.id          = interfaceId
        pyutInterface.description = description

        implementors: Element = interface.get_elements('Implementor')
        for implementor in implementors:
            pyutInterface.addImplementor(implementor['implementingClassName'])

        pyutInterface.methods = self._interfaceMethodsToPyutMethods(interface=interface)
        return pyutInterface

    def actorToPyutActor(self, graphicActor: Element) -> PyutActor:
        """
        <GraphicActor width="87" height="114" x="293" y="236">
            <Actor id="1" name="BasicActor" filename=""/>
        </GraphicActor>

        Args:
            graphicActor:   untangle Element in the above format

        Returns:   PyutActor
        """
        actorElement: Element   = graphicActor.Actor
        pyutActor:    PyutActor = PyutActor()

        pyutActor = cast(PyutActor, self._addPyutObjectAttributes(pyutElement=actorElement, pyutObject=pyutActor))

        return pyutActor

    def useCaseToPyutUseCase(self, graphicUseCase: Element) -> PyutUseCase:
        """
        <GraphicUseCase width="100" height="60" x="575" y="250">
            <UseCase id="2" name="Basic Use Case" filename=""/>
        </GraphicUseCase>

        Args:
            graphicUseCase:  An untangle Element in the above format

        Returns:  PyutUseCase
        """
        useCaseElement: Element     = graphicUseCase.UseCase
        pyutUseCase:    PyutUseCase = PyutUseCase()

        pyutUseCase = cast(PyutUseCase, self._addPyutObjectAttributes(pyutElement=useCaseElement, pyutObject=pyutUseCase))

        return pyutUseCase

    def linkToPyutLink(self, singleLink: Element, source: PyutClass, destination: PyutClass) -> PyutLink:
        linkTypeStr:     str          = singleLink['type']
        linkType:        PyutLinkType = PyutLinkType.toEnum(linkTypeStr)
        cardSrc:         str          = singleLink['cardSrc']
        cardDest:        str          = singleLink['cardDestination']
        bidir:           bool         = str2bool(singleLink['bidir'])
        linkDescription: str          = singleLink['name']

        pyutLink: PyutLink = PyutLink(name=linkDescription,
                                      linkType=linkType,
                                      cardSrc=cardSrc, cardDest=cardDest,
                                      bidir=bidir,
                                      source=source,
                                      destination=destination)

        return pyutLink

    def sdInstanceToPyutSDInstance(self, graphicSDInstance: Element) -> PyutSDInstance:

        instanceElement: Element        = graphicSDInstance.SDInstance
        pyutSDInstance:  PyutSDInstance = PyutSDInstance()

        pyutSDInstance.id                     = int(instanceElement['id'])
        pyutSDInstance.instanceName           = instanceElement['instanceName']
        pyutSDInstance.instanceLifeLineLength = secureInteger(instanceElement['lifeLineLength'])

        return pyutSDInstance

    def sdMessageToPyutSDMessage(self, graphicSDMessage: Element) -> ConvolutedPyutSDMessageInformation:
        """
        TODO:  Need to fix how SD Messages are created
        Args:
            graphicSDMessage:

        Returns:  Bogus data class
        """

        messageElement: Element       = graphicSDMessage.SDMessage
        pyutSDMessage:  PyutSDMessage = PyutSDMessage()

        pyutSDMessage.id = int(messageElement['id'])
        pyutSDMessage.setMessage(messageElement['message'])
        pyutSDMessage.linkType = PyutLinkType.SD_MESSAGE

        srcID: int = int(messageElement['srcID'])
        dstID: int = int(messageElement['dstID'])

        srcTime: int = int(messageElement['srcTime'])
        dstTime: int = int(messageElement['dstTime'])

        pyutSDMessage.setSrcTime(value=srcTime, updateOGLObject=False)
        pyutSDMessage.setDstTime(value=dstTime, updateOGLObject=False)

        bogus: ConvolutedPyutSDMessageInformation = ConvolutedPyutSDMessageInformation()

        bogus.pyutSDMessage = pyutSDMessage
        bogus.sourceId      = srcID
        bogus.destinationId = dstID

        return bogus

    def _methodToPyutMethods(self, classElement: Element) -> UntangledPyutMethods:
        """
        The pyutClass may not have methods;
        Args:
            classElement:  The pyutClassElement

        Returns:  May return an empty list
        """
        untangledPyutMethods: UntangledPyutMethods = UntangledPyutMethods([])

        methodElements = classElement.get_elements('Method')
        for methodElement in methodElements:
            methodName: str                = methodElement['name']
            visibility: PyutVisibilityEnum = PyutVisibilityEnum.toEnum(methodElement['visibility'])
            self.logger.debug(f"{methodName=} - {visibility=}")

            pyutMethod: PyutMethod = PyutMethod(name=methodName, visibility=visibility)

            pyutMethod.modifiers = self._modifierToPyutMethodModifiers(methodElement=methodElement)

            returnElement = methodElement.get_elements('Return')

            if len(returnElement) > 0:
                pyutType: PyutType = PyutType(value=returnElement[0]['type'])
                pyutMethod.returnType = pyutType

            parameters = self._paramToPyutParameters(methodElement)
            pyutMethod.parameters = parameters
            pyutMethod.sourceCode = self._sourceCodeToPyutSourceCode(methodElement=methodElement)

            untangledPyutMethods.append(pyutMethod)

        return untangledPyutMethods

    def _fieldToPyutFields(self, classElement: Element) -> UntangledPyutFields:
        untangledPyutFields: UntangledPyutFields = UntangledPyutFields([])

        fieldElements: Elements = classElement.get_elements('Field')
        for fieldElement in fieldElements:
            visibility:    PyutVisibilityEnum = PyutVisibilityEnum.toEnum(fieldElement['visibility'])
            paramElements: Elements           = fieldElement.get_elements('Param')
            assert len(paramElements) == 1, 'Curiously there should be only one'

            paramElement: Element = paramElements[0]
            fieldName: str = paramElement['name']
            pyutType:  PyutType  = PyutType(paramElement['type'])

            pyutField: PyutField = PyutField(name=fieldName, visibility=visibility, fieldType=pyutType)

            untangledPyutFields.append(pyutField)

        return untangledPyutFields

    def _modifierToPyutMethodModifiers(self, methodElement: Element) -> PyutModifiers:
        """
        Should be in this form:

        <Modifier name="modifier1,modifier2,modifier3"/>

        Args:
            methodElement:

        Returns:  PyutModifiers if not exist returns an empty
        """

        modifierElements = methodElement.get_elements('Modifier')

        pyutModifiers: PyutModifiers = PyutModifiers([])
        if len(modifierElements) > 0:
            modifierElement: Element   = modifierElements[0]
            names:           str       = modifierElement['name']    # comma delimited string
            nameList:        List[str] = names.split(',')
            for modifierName in nameList:
                pyutModifier: PyutModifier = PyutModifier(modifierTypeName=modifierName)
                pyutModifiers.append(pyutModifier)

        return pyutModifiers

    def _paramToPyutParameters(self, methodElement: Element) -> PyutParameters:

        parameterElements = methodElement.get_elements('Param')     # TODO:  https://github.com/hasii2011/PyUt/issues/326
        untangledPyutMethodParameters: PyutParameters = PyutParameters([])
        for parameterElement in parameterElements:
            name:           str = parameterElement['name']
            defaultValue:   str = parameterElement['defaultValue']
            parameterType:  PyutType = PyutType(parameterElement['type'])

            pyutParameter: PyutParameter = PyutParameter(name=name, parameterType=parameterType, defaultValue=defaultValue)
            # <Param name="intParameter" type="int" defaultValue="0"/>
            # <Param name="floatParameter" type="float" defaultValue="0.0"/>
            # <Param name="stringParameter" type="str" defaultValue="''"/>
            untangledPyutMethodParameters.append(pyutParameter)

        return untangledPyutMethodParameters

    def _sourceCodeToPyutSourceCode(self, methodElement: Element) -> SourceCode:

        sourceCodeElements = methodElement.get_elements('SourceCode')
        codeElements = sourceCodeElements[0].get_elements('Code')
        sourceCode: SourceCode = SourceCode([])
        for codeElement in codeElements:
            self.logger.debug(f'{codeElement.cdata=}')
            codeLine: str = codeElement.cdata
            sourceCode.append(codeLine)
        return sourceCode

    def _interfaceMethodsToPyutMethods(self, interface: Element) -> List[PyutMethod]:

        pyutMethods: List[PyutMethod] = self._methodToPyutMethods(interface)

        return pyutMethods

    def _addPyutObjectAttributes(self, pyutElement: Element, pyutObject: PyutObject) -> PyutObject:
        """

        Args:
            pyutElement:    pyutElement XML with common keys
            pyutObject:     The PyutObject to update

        Returns:  The updated pyutObject as
        """

        pyutObject.id       = int(pyutElement['id'])    # TODO revisit this when we start using UUIDs
        pyutObject.name     = pyutElement['name']
        pyutObject.fileName = pyutElement['fileName']

        return pyutObject




from typing import NewType
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass
from dataclasses import field

from untangle import parse
from untangle import Element

from ogl.OglClass import OglClass
from ogl.OglNote import OglNote
from ogl.OglText import OglText

from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutNote import PyutNote
from pyutmodel.PyutText import PyutText

from untanglepyut.BaseUnTangle import BaseUnTangle

from untanglepyut.Common import Elements
from untanglepyut.Common import toGraphicInfo
from untanglepyut.Common import GraphicInformation
from untanglepyut.Common import UntangledOglLinks
from untanglepyut.Common import createUntangledOglActors
from untanglepyut.Common import createUntangledOglLinks
from untanglepyut.Common import createUntangledOglUseCases

from untanglepyut.Types import UntangledOglActors
from untanglepyut.Types import UntangledOglClasses
from untanglepyut.Types import UntangledOglNotes
from untanglepyut.Types import UntangledOglTexts
from untanglepyut.Types import UntangledOglUseCases

from untanglepyut.UnTangleOglLinks import LinkableOglObjects
from untanglepyut.UnTangleOglLinks import createLinkableOglObjects
from untanglepyut.UnTanglePyut import UnTanglePyut
from untanglepyut.UnTangleOglLinks import UnTangleOglLinks
from untanglepyut.UnTangleUseCaseDiagram import UnTangleUseCaseDiagram
from untanglepyut.UntangleSequenceDiagram import OglSDMessages
from untanglepyut.UntangleSequenceDiagram import OglSDInstances
from untanglepyut.UntangleSequenceDiagram import UntangleSequenceDiagram
from untanglepyut.UntangleSequenceDiagram import createOglSDMessages
from untanglepyut.UntangleSequenceDiagram import createOglSDInstances


@dataclass
class ProjectInformation:
    version:  str = cast(str, None)
    codePath: str = cast(str, None)


"""
Factory methods for our dataclasses
"""


def createUntangledOglClasses() -> UntangledOglClasses:
    """
    Factory method to create  the UntangledClasses data structure;

    Returns:  A new data structure
    """
    return UntangledOglClasses([])


def createUntangledOglNotes() -> UntangledOglNotes:
    return UntangledOglNotes([])


def createUntangledOglTexts() -> UntangledOglTexts:
    return UntangledOglTexts([])


@dataclass
class Document:
    """
    Create a UseCaseDocument and a ClassDiagramDocument
    """
    documentType:    str = ''
    documentTitle:   str = ''
    scrollPositionX: int = -1
    scrollPositionY: int = -1
    pixelsPerUnitX:  int = -1
    pixelsPerUnitY:  int = -1
    oglClasses:      UntangledOglClasses  = field(default_factory=createUntangledOglClasses)
    oglLinks:        UntangledOglLinks    = field(default_factory=createUntangledOglLinks)
    oglNotes:        UntangledOglNotes    = field(default_factory=createUntangledOglNotes)
    oglTexts:        UntangledOglTexts    = field(default_factory=createUntangledOglTexts)
    oglActors:       UntangledOglActors   = field(default_factory=createUntangledOglActors)
    oglUseCases:     UntangledOglUseCases = field(default_factory=createUntangledOglUseCases)
    oglSDInstances:  OglSDInstances = field(default_factory=createOglSDInstances)
    oglSDMessages:   OglSDMessages  = field(default_factory=createOglSDMessages)


# @dataclass
# class SDDocument(Document):
#     oglSDInstances:  OglSDInstances = field(default_factory=createOglSDInstances)
#     oglSDMessages:   OGLSDMessages  = field(default_factory=createOGLSDMessages)


DocumentTitle = NewType('DocumentTitle', str)
Documents     = NewType('Documents', dict[DocumentTitle, Document])


class UnTangler(BaseUnTangle):

    def __init__(self):
        """
        """
        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._projectInformation: ProjectInformation = ProjectInformation()
        self._documents:          Documents          = Documents({})

        self._untanglePyut:     UnTanglePyut     = UnTanglePyut()
        self._untangleOglLinks: UnTangleOglLinks = UnTangleOglLinks()

    @property
    def projectInformation(self) -> ProjectInformation:
        """
        This property return nothing valid until you untangle the file

        Returns:  The project information of the untangled pyut file
        """
        return self._projectInformation

    @property
    def documents(self) -> Documents:
        return self._documents

    def untangleFile(self, fqFileName: str):
        """
        Read input file and untangle to Ogl
        Args:
            fqFileName:  The file name with the XML

        """
        xmlString:   str     = self.getRawXml(fqFileName=fqFileName)
        self.untangleXml(xmlString=xmlString)

    def untangleXml(self, xmlString: str):
        """
        Untangle the input Xml string to Ogl
        Args:
            xmlString: The string wit the raw XML
        """
        root:        Element = parse(xmlString)
        pyutProject: Element = root.PyutProject

        self._populateProjectInformation(pyutProject=pyutProject)

        for pyutDocument in pyutProject.PyutDocument:
            document: Document = self._updateCurrentDocumentInformation(pyutDocument=pyutDocument)

            self._documents[DocumentTitle(document.documentTitle)] = document

            self.logger.debug(f'{document=}')
            if document.documentType == 'CLASS_DIAGRAM':
                document.oglClasses = self._graphicClassesToOglClasses(pyutDocument=pyutDocument)
                document.oglNotes   = self._graphicNotesToOglNotes(pyutDocument=pyutDocument)
                document.oglTexts   = self._graphicalTextToOglTexts(pyutDocument=pyutDocument)

                linkableOglObjects: LinkableOglObjects = self._buildDictionary(document=document)
                document.oglLinks   = self._untangleOglLinks.graphicLinksToOglLinks(pyutDocument, linkableOglObjects=linkableOglObjects)
            elif document.documentType == 'SEQUENCE_DIAGRAM':
                untangleSequenceDiagram: UntangleSequenceDiagram = UntangleSequenceDiagram()

                untangleSequenceDiagram.unTangle(pyutDocument=pyutDocument)
                document.oglSDInstances = untangleSequenceDiagram.oglSDInstances
                document.oglSDMessages  = untangleSequenceDiagram.oglSDMessages

            elif document.documentType == 'USECASE_DIAGRAM':

                unTangleUseCaseDiagram: UnTangleUseCaseDiagram = UnTangleUseCaseDiagram()

                unTangleUseCaseDiagram.unTangle(pyutDocument=pyutDocument)
                document.oglActors   = unTangleUseCaseDiagram.oglActors
                document.oglUseCases = unTangleUseCaseDiagram.oglUseCases

                linkableOglObjects = self._buildDictionary(document=document)
                document.oglLinks  = self._untangleOglLinks.graphicLinksToOglLinks(pyutDocument, linkableOglObjects=linkableOglObjects)
            else:
                assert False, f'Unknown document type: f{document.documentType}'

    def getRawXml(self, fqFileName: str) -> str:
        """
        Convenience method to read a file.  Assumes the file has XML.
        No check is done to verify this
        Args:
            fqFileName: The file to read

        Returns:  The contents of the file
        """
        try:
            with open(fqFileName, "r") as xmlFile:
                xmlString: str = xmlFile.read()
        except (ValueError, Exception) as e:
            self.logger.error(f'xml open:  {e}')
            raise e

        return xmlString

    def _populateProjectInformation(self, pyutProject: Element):
        self._projectInformation.version  = pyutProject['version']
        self._projectInformation.codePath = pyutProject['CodePath']

    def _updateCurrentDocumentInformation(self, pyutDocument: Element) -> Document:

        documentInformation: Document = Document()

        documentTitle: DocumentTitle = DocumentTitle(pyutDocument['title'])

        documentInformation.documentType = pyutDocument['type']
        documentInformation.documentTitle = documentTitle

        documentInformation.scrollPositionX = int(pyutDocument['scrollPositionX'])
        documentInformation.scrollPositionY = int(pyutDocument['scrollPositionY'])
        documentInformation.pixelsPerUnitX  = int(pyutDocument['pixelsPerUnitX'])
        documentInformation.pixelsPerUnitY  = int(pyutDocument['pixelsPerUnitY'])

        self.logger.debug(f'{documentInformation=}')

        return documentInformation

    def _graphicClassesToOglClasses(self, pyutDocument: Element) -> UntangledOglClasses:

        oglClasses:     UntangledOglClasses = createUntangledOglClasses()
        graphicClasses: Elements            = pyutDocument.get_elements('GraphicClass')
        
        for graphicClass in graphicClasses:
            self.logger.debug(f'{graphicClass=}')

            graphicInformation: GraphicInformation = toGraphicInfo(graphicElement=graphicClass)
            oglClass: OglClass = OglClass(w=graphicInformation.width, h=graphicInformation.height)
            oglClass.SetPosition(x=graphicInformation.x, y=graphicInformation.y)
            #
            # This is necessary if it is never added to a diagram
            # and immediately serialized
            #
            self._updateModel(oglObject=oglClass, graphicInformation=graphicInformation)

            pyutClass: PyutClass = self._untanglePyut.classToPyutClass(graphicClass=graphicClass)
            oglClass.pyutObject = pyutClass
            oglClasses.append(oglClass)

        return oglClasses

    def _graphicNotesToOglNotes(self, pyutDocument: Element) -> UntangledOglNotes:
        """

        Args:
            pyutDocument:

        Returns: untangled OglNote objects if any exist, else an empty list
        """
        oglNotes:     UntangledOglNotes = createUntangledOglNotes()
        graphicNotes: Elements          = pyutDocument.get_elements('GraphicNote')

        for graphicNote in graphicNotes:
            self.logger.debug(f'{graphicNote}')

            graphicInformation: GraphicInformation = toGraphicInfo(graphicElement=graphicNote)
            oglNote:            OglNote            = OglNote(w=graphicInformation.width, h=graphicInformation.height)
            oglNote.SetPosition(x=graphicInformation.x, y=graphicInformation.y)
            self._updateModel(oglObject=oglNote, graphicInformation=graphicInformation)
            pyutNote: PyutNote = self._untanglePyut.noteToPyutNote(graphicNote=graphicNote)
            oglNote.pyutObject = pyutNote
            oglNotes.append(oglNote)

        return oglNotes

    def _graphicalTextToOglTexts(self, pyutDocument: Element) -> UntangledOglTexts:
        """
        Yeah, yeah, I know bad English;

        Args:
            pyutDocument:  The Element document

        Returns:  untangled OglText objects if any exist, else an empty list
        """
        oglTexts:     UntangledOglTexts = createUntangledOglTexts()
        graphicTexts: Elements          = pyutDocument.get_elements('GraphicText')

        for graphicText in graphicTexts:
            self.logger.debug(f'{graphicText}')

            graphicInformation: GraphicInformation = toGraphicInfo(graphicElement=graphicText)
            pyutText:           PyutText           = self._untanglePyut.textToPyutText(graphicText=graphicText)
            oglText:            OglText            = OglText(pyutText=pyutText, width=graphicInformation.width, height=graphicInformation.height)
            oglText.SetPosition(x=graphicInformation.x, y=graphicInformation.y)
            #
            # This is necessary if it is never added to a diagram
            # and immediately serialized
            #
            self._updateModel(oglObject=oglText, graphicInformation=graphicInformation)
            oglText.pyutText = pyutText
            oglTexts.append(oglText)

        return oglTexts

    def _buildDictionary(self, document: Document) -> LinkableOglObjects:
        """

        Args:
            document:   The created document either Use case or class diagram

        Returns:  Linkable Objects Dictionary
        """

        linkableOglObjects: LinkableOglObjects = createLinkableOglObjects()

        for oglClass in document.oglClasses:
            linkableOglObjects[oglClass.pyutObject.id] = oglClass

        for oglNote in document.oglNotes:
            linkableOglObjects[oglNote.pyutObject.id] = oglNote

        for oglUseCase in document.oglUseCases:
            linkableOglObjects[oglUseCase.pyutObject.id] = oglUseCase

        for oglActor in document.oglActors:
            linkableOglObjects[oglActor.pyutObject.id] = oglActor

        return linkableOglObjects



from typing import Dict
from typing import List
from typing import NewType

from logging import Logger
from logging import getLogger

from untangle import Element

from pyutmodel.PyutSDInstance import PyutSDInstance
from pyutmodel.PyutSDMessage import PyutSDMessage

from ogl.sd.OglSDInstance import OglSDInstance
from ogl.sd.OglSDMessage import OglSDMessage

from untanglepyut.Types import OglSDInstances
from untanglepyut.Types import OglSDMessages

from untanglepyut.BaseUnTangle import BaseUnTangle

from untanglepyut.Common import GraphicInformation
from untanglepyut.Common import toGraphicInfo

from untanglepyut.UnTanglePyut import ConvolutedPyutSDMessageInformation
from untanglepyut.UnTanglePyut import UnTanglePyut


def createOglSDInstances() -> OglSDInstances:
    return OglSDInstances({})


def createOglSDMessages() -> OglSDMessages:
    return OglSDMessages({})


class UntangleSequenceDiagram(BaseUnTangle):

    def __init__(self):

        super().__init__()

        self.logger: Logger = getLogger(__name__)

        self._oglSDInstances: OglSDInstances = createOglSDInstances()
        self._oglSDMessages:  OglSDMessages  = createOglSDMessages()

        self._untanglePyut: UnTanglePyut = UnTanglePyut()

    def unTangle(self, pyutDocument: Element):
        """

        Args:
            pyutDocument:  The pyut untangle element that represents a sequence diagram
        """
        self._oglSDInstances = self._untangleSDInstances(pyutDocument=pyutDocument)
        self._oglSDMessages  = self._untangleSDMessages(pyutDocument=pyutDocument)

    @property
    def oglSDInstances(self) -> OglSDInstances:
        return self._oglSDInstances

    @property
    def oglSDMessages(self) -> OglSDMessages:
        return self._oglSDMessages

    def _untangleSDInstances(self, pyutDocument: Element) -> OglSDInstances:

        oglSDInstances:     OglSDInstances = createOglSDInstances()
        graphicSDInstances: List[Element]   = pyutDocument.get_elements('GraphicSDInstance')

        for graphicSDInstance in graphicSDInstances:
            self.logger.debug(f'{graphicSDInstance=}')
            pyutSDInstance: PyutSDInstance     = self._untanglePyut.sdInstanceToPyutSDInstance(graphicSDInstance=graphicSDInstance)

            oglSDInstance:  OglSDInstance      = OglSDInstance(pyutSDInstance)
            graphicInfo:    GraphicInformation = toGraphicInfo(graphicElement=graphicSDInstance)

            oglSDInstance.SetSize(width=graphicInfo.width, height=graphicInfo.height)
            oglSDInstance.SetPosition(x=graphicInfo.x, y=graphicInfo.y)

            self._updateModel(oglObject=oglSDInstance, graphicInformation=graphicInfo)

            oglSDInstances[pyutSDInstance.id] = oglSDInstance
        return oglSDInstances

    def _untangleSDMessages(self, pyutDocument: Element) -> OglSDMessages:

        oglSDMessages:     OglSDMessages = createOglSDMessages()
        graphicSDMessages: List[Element] = pyutDocument.get_elements('GraphicSDMessage')

        for graphicSDMessage in graphicSDMessages:
            bogus: ConvolutedPyutSDMessageInformation = self._untanglePyut.sdMessageToPyutSDMessage(graphicSDMessage=graphicSDMessage)

            pyutSDMessage: PyutSDMessage = bogus.pyutSDMessage

            srcInstance: OglSDInstance = self._oglSDInstances[bogus.sourceId]
            dstInstance: OglSDInstance = self._oglSDInstances[bogus.destinationId]

            pyutSDMessage.setSource(srcInstance.pyutObject)          # Ugh, time was set by sdMessageToPyutSDMessage
            pyutSDMessage.setDestination(dstInstance.pyutObject)     # This "split" functionality must be fixed
            oglSDMessage: OglSDMessage = OglSDMessage(srcShape=srcInstance, pyutSDMessage=pyutSDMessage, dstShape=dstInstance)

            srcInstance.addLink(link=oglSDMessage)
            dstInstance.addLink(link=oglSDMessage)

            oglSDMessages[pyutSDMessage.id] = oglSDMessage

        return oglSDMessages




from typing import Dict
from typing import List
from typing import NewType
from typing import Union

from ogl.OglActor import OglActor
from ogl.OglClass import OglClass
from ogl.OglLink import OglLink
from ogl.OglNote import OglNote
from ogl.OglText import OglText
from ogl.OglUseCase import OglUseCase
from ogl.OglInterface2 import OglInterface2
from ogl.sd.OglSDInstance import OglSDInstance
from ogl.sd.OglSDMessage import OglSDMessage

UntangledOglClasses    = NewType('UntangledOglClasses', List[OglClass])
UntangledLinks         = Union[OglLink, OglInterface2]

UntangledOglLinks      = NewType('UntangledOglLinks',    List[UntangledLinks])
UntangledOglNotes      = NewType('UntangledOglNotes',    List[OglNote])
UntangledOglTexts      = NewType('UntangledOglTexts',    List[OglText])
UntangledOglActors     = NewType('UntangledOglActors',   List[OglActor])
UntangledOglUseCases   = NewType('UntangledOglUseCases', List[OglUseCase])

OglSDInstances = NewType('OglSDInstances', Dict[int, OglSDInstance])
OglSDMessages  = NewType('OglSDMessages',  Dict[int, OglSDMessage])




from logging import Logger
from logging import getLogger

from miniogl.ShapeModel import ShapeModel
from ogl.OglObject import OglObject

from untanglepyut.Common import GraphicInformation


class BaseUnTangle:

    def __init__(self):
        self.baseLogger: Logger = getLogger(__name__)

    def _updateModel(self, oglObject: OglObject, graphicInformation: GraphicInformation) -> ShapeModel:
        """
        This is necessary if it is never added to a diagram
        and immediately serialized

        Args:
            oglObject:      OglObject with a model
            graphicInformation:   The graphic class graphic information

        Returns:  The updated shape model as a way of documenting that we updated it
        """
        model: ShapeModel = oglObject.GetModel()
        model.SetPosition(x=graphicInformation.x, y=graphicInformation.y)

        return model



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()
LICENSE = (HERE / 'LICENSE').read_text()

setup(
    name="untanglepyut",
    version="0.5.1",
    author_email='Humberto.A.Sanchez.II@gmail.com',
    description='XML to Ogl Object Model',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/hasii2011/untanglepyut",
    packages=[
        'untanglepyut'
    ],
    package_data={
        'untanglepyut': ['py.typed'],
    },

    install_requires=['ogl==0.60.5', 'untangle==1.2.1'],
)




from logging import Logger
from logging import getLogger
from typing import Dict
from typing import List
from typing import NewType
from typing import Union
from typing import cast

from dataclasses import dataclass

from pyutmodel.PyutClass import PyutClass
from untangle import Element

from miniogl.AttachmentLocation import AttachmentLocation
from miniogl.ControlPoint import ControlPoint
from miniogl.SelectAnchorPoint import SelectAnchorPoint

from pyutmodel.PyutLinkType import PyutLinkType
from pyutmodel.PyutInterface import PyutInterface
from pyutmodel.PyutLink import PyutLink

from ogl.OglPosition import OglPosition
from ogl.OglAggregation import OglAggregation
from ogl.OglAssociation import OglAssociation
from ogl.OglComposition import OglComposition
from ogl.OglInheritance import OglInheritance
from ogl.OglInterface import OglInterface
from ogl.OglNoteLink import OglNoteLink
from ogl.OglNote import OglNote
from ogl.OglActor import OglActor
from ogl.OglUseCase import OglUseCase
from ogl.OglClass import OglClass
from ogl.OglLink import OglLink
from ogl.OglInterface2 import OglInterface2
from ogl.OglAssociationLabel import OglAssociationLabel

from untanglepyut.Common import UntangledControlPoints
from untanglepyut.Common import UntangledOglLinks
from untanglepyut.Common import createUntangledOglLinks
from untanglepyut.Common import str2bool

from untanglepyut.UnTanglePyut import UnTanglePyut


LinkableOglObject = Union[OglClass, OglNote, OglActor, OglUseCase]

LinkableOglObjects = NewType('LinkableOglObjects',   Dict[int, LinkableOglObject])


def createLinkableOglObjects() -> LinkableOglObjects:
    return LinkableOglObjects({})


@dataclass
class GraphicLinkAttributes:

    srcX:   int = -1
    srcY:   int = -1
    dstX:   int = -1
    dstY:   int = -1
    spline: bool = False


def fromGraphicLink(graphicLink: Element) -> GraphicLinkAttributes:

    gla: GraphicLinkAttributes = GraphicLinkAttributes()
    gla.srcX = int(graphicLink['srcX'])
    gla.srcY = int(graphicLink['srcY'])
    gla.dstX = int(graphicLink['dstX'])
    gla.dstY = int(graphicLink['dstY'])

    gla.spline = str2bool(graphicLink['spline'])

    return gla


class UnTangleOglLinks:
    """
    Currently, unsupported:

    ```html
      <LabelCenter x="579" y="300"/>
      <LabelSrc x="579" y="300"/>
      <LabelDst x="579" y="300"/>
    ```
    See:
    """

    def __init__(self):

        self.logger: Logger = getLogger(__name__)

        self._untanglePyut:     UnTanglePyut     = UnTanglePyut()

    def graphicLinksToOglLinks(self, pyutDocument: Element, linkableOglObjects: LinkableOglObjects) -> UntangledOglLinks:
        """
        Convert from XML to Ogl Links

        Args:
            pyutDocument:  The Element that represents the Class Diagram XML
            linkableOglObjects:    OGL objects that can have links

        Returns:  The links between any of the above objects.  Also returns the graphic lollipop links
        """

        oglLinks: UntangledOglLinks = createUntangledOglLinks()

        graphicLinks: Element = pyutDocument.get_elements('GraphicLink')
        for graphicLink in graphicLinks:
            oglLink: OglLink = self._graphicLinkToOglLink(graphicLink, linkableOglObjects=linkableOglObjects)
            oglLinks.append(oglLink)

        graphicLollipops: Element = pyutDocument.get_elements('GraphicLollipop')
        for graphicLollipop in graphicLollipops:
            oglInterface2: OglInterface2 = self._graphicLollipopToOglInterface(graphicLollipop, linkableOglObjects)
            oglLinks.append(oglInterface2)

        return oglLinks

    def _graphicLinkToOglLink(self, graphicLink: Element, linkableOglObjects: LinkableOglObjects) -> OglLink:
        """
        This code is way too convoluted.  Failing to do any of these step in this code leads to BAD
        visual representations.
        TODO:  Figure out how to simplify this code and/or make it more readable and obvious on how to create
        links (of whatever kind) between 2 OglClass'es

        Args:
            graphicLink:        The XML `GraphicClass` element
            linkableOglObjects:    OGL objects that can have links

        Returns:  A fully formed OglLink including control points
        """

        assert len(linkableOglObjects) != 0, 'Developer forgot to create dictionary'
        gla: GraphicLinkAttributes = fromGraphicLink(graphicLink=graphicLink)

        links: Element = graphicLink.get_elements('Link')
        assert len(links) == 1, 'Should only ever be one'

        singleLink:  Element = links[0]
        sourceId:    int = int(singleLink['sourceId'])
        dstId:       int = int(singleLink['destId'])
        self.logger.debug(f'graphicLink= {gla.srcX=} {gla.srcY=} {gla.dstX=} {gla.dstY=} {gla.spline=}')

        try:
            srcShape: LinkableOglObject = linkableOglObjects[sourceId]
            dstShape: LinkableOglObject = linkableOglObjects[dstId]
        except KeyError as ke:
            self.logger.error(f'Developer Error -- srcId: {sourceId} - dstId: {dstId}  KeyError index: {ke}')
            return cast(OglLink, None)

        pyutLink: PyutLink = self._untanglePyut.linkToPyutLink(singleLink, source=srcShape.pyutObject, destination=dstShape.pyutObject)
        oglLink:  OglLink  = self._oglLinkFactory(srcShape=srcShape, pyutLink=pyutLink, destShape=dstShape,
                                                  linkType=pyutLink.linkType,
                                                  srcPos=(gla.srcX, gla.srcY),
                                                  dstPos=(gla.dstX, gla.dstY)
                                                  )
        oglLink.SetSpline(gla.spline)
        srcShape.addLink(oglLink)
        dstShape.addLink(oglLink)

        # put the anchors at the right position
        srcAnchor = oglLink.GetSource()
        dstAnchor = oglLink.GetDestination()
        srcAnchor.SetPosition(gla.srcX, gla.srcY)
        dstAnchor.SetPosition(gla.dstX, gla.dstY)

        srcModel = srcAnchor.GetModel()
        srcModel.SetPosition(x=gla.srcX, y=gla.srcY)
        dstModel = dstAnchor.GetModel()
        dstModel.SetPosition(x=gla.dstX, y=gla.dstY)

        # add the control points to the line
        line   = srcAnchor.GetLines()[0]     # only 1 line per anchor in Pyut
        parent = line.GetSource().GetParent()
        selfLink: bool = parent is oglLink.GetDestination().GetParent()

        controlPoints: UntangledControlPoints = self._generateControlPoints(graphicLink=graphicLink)
        for controlPoint in controlPoints:
            oglLink.AddControl(controlPoint)
            if selfLink:
                x, y = controlPoint.GetPosition()
                controlPoint.SetParent(parent)
                controlPoint.SetPosition(x, y)

        if isinstance(oglLink, OglAssociation):
            self.__furtherCustomizeAssociationLink(graphicLink, oglLink)

        self._reconstituteLinkDataModel(oglLink)

        return oglLink

    def _oglLinkFactory(self, srcShape, pyutLink, destShape, linkType: PyutLinkType, srcPos=None, dstPos=None):
        """
        Used to get a OglLink of the given linkType.

        Args:
            srcShape:   Source shape
            pyutLink:   Conceptual links associated with the graphical links.
            destShape:  Destination shape
            linkType:   The linkType of the link (OGL_INHERITANCE, ...)
            srcPos:     source position
            dstPos:     destination position

        Returns:  The requested link
        """
        if linkType == PyutLinkType.AGGREGATION:
            return OglAggregation(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)

        elif linkType == PyutLinkType.COMPOSITION:
            return OglComposition(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)

        elif linkType == PyutLinkType.INHERITANCE:
            return OglInheritance(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.ASSOCIATION:
            return OglAssociation(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)

        elif linkType == PyutLinkType.INTERFACE:
            return OglInterface(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)

        elif linkType == PyutLinkType.NOTELINK:
            return OglNoteLink(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.SD_MESSAGE:
            assert False, 'Sequence Diagram Messages not supported'
            # return OglSDMessage(srcShape=srcShape, pyutSDMessage=pyutLink, dstShape=destShape)
        else:
            self.logger.error(f"Unknown OglLinkType: {linkType}")
            return None

    def _graphicLollipopToOglInterface(self, graphicLollipop: Element, linkableOglObjects: LinkableOglObjects) -> OglInterface2:

        assert len(linkableOglObjects) != 0, 'Developer forgot to create dictionary'

        #
        # TODO: Do not use these x,y positions;  They are diagram relative
        #
        # x: int = int(graphicLollipop['x'])
        # y: int = int(graphicLollipop['y'])
        attachmentLocationStr: str                = graphicLollipop['attachmentPoint']
        attachmentLocation:    AttachmentLocation = AttachmentLocation.toEnum(attachmentLocationStr)

        elements: Element = graphicLollipop.get_elements('Interface')
        assert len(elements) == 1, 'If more than one interface tag the XML is invalid'
        interfaceElement: Element        = elements[0]
        pyutInterface:    PyutInterface = self._untanglePyut.interfaceToPyutInterface(interface=interfaceElement)

        oglClass:    OglClass    = self._getOglClassFromName(pyutInterface.implementors[0], linkableOglObjects)
        oglPosition: OglPosition = self._determineAttachmentPoint(attachmentLocation, oglClass)
        self.logger.debug(f'{oglPosition.x=},{oglPosition.y=}')

        anchorPoint:      SelectAnchorPoint = SelectAnchorPoint(x=oglPosition.x, y=oglPosition.y, attachmentPoint=attachmentLocation, parent=oglClass)
        oglInterface2:    OglInterface2     = OglInterface2(pyutInterface=pyutInterface, destinationAnchor=anchorPoint)

        return oglInterface2

    def _getOglClassFromName(self, className: str, linkableOglObjects: LinkableOglObjects) -> OglClass:
        """
        Looks up a name in the linkable objects dictionary and return the associated class
        TODO: Make a simple lookup and catch any Key errors

        Args:
            className:
            linkableOglObjects:

        Returns:
        """

        foundClass: OglClass = cast(OglClass, None)
        for oglClass in linkableOglObjects.values():
            if oglClass.pyutObject.name == className:
                foundClass = cast(OglClass, oglClass)
                break
        assert foundClass is not None, 'XML must be in error'
        return foundClass

    def _determineAttachmentPoint(self, attachmentPoint: AttachmentLocation, oglClass: OglClass) -> OglPosition:
        """
        Even though we serialize the attachment point location that position is relative to the diagram.
        When we recreate the attachment point position we have to create it relative to its parent
        TODO: When the Pyut serializer makes the positions relative to the implementor we will not need this code

        Args:
            attachmentPoint:    Where on the parent
            oglClass:           The implementor

        Returns:  An OglPosition with coordinates relative to the implementor
        """

        oglPosition: OglPosition = OglPosition()

        dw, dh     = oglClass.GetSize()

        if attachmentPoint == AttachmentLocation.NORTH:
            northX: int = dw // 2
            northY: int = 0
            oglPosition.x = northX
            oglPosition.y = northY
        elif attachmentPoint == AttachmentLocation.SOUTH:
            southX = dw // 2
            southY = dh
            oglPosition.x = southX
            oglPosition.y = southY
        elif attachmentPoint == AttachmentLocation.WEST:
            westX: int = 0
            westY: int = dh // 2
            oglPosition.x = westX
            oglPosition.y = westY
        elif attachmentPoint == AttachmentLocation.EAST:
            eastX: int = dw
            eastY: int = dh // 2
            oglPosition.x = eastX
            oglPosition.y = eastY
        else:
            self.logger.warning(f'Unknown attachment point: {attachmentPoint}')
            assert False, 'Unknown attachment point'

        return oglPosition

    def _generateControlPoints(self, graphicLink: Element) -> UntangledControlPoints:

        controlPoints: UntangledControlPoints = UntangledControlPoints([])

        controlPointElements: Element = graphicLink.get_elements('ControlPoint')
        for controlPointElement in controlPointElements:
            x: int = int(controlPointElement['x'])
            y: int = int(controlPointElement['y'])
            controlPoint: ControlPoint = ControlPoint(x=x, y=y)
            controlPoints.append(controlPoint)

        return controlPoints

    def _reconstituteLinkDataModel(self, oglLink: OglLink):
        """
        Updates one the following lists in a PyutLinkedObject:

        ._parents   for Inheritance links
        ._links     for all other link types

        Args:
            oglLink:       An OglLink
        """
        srcShape:  OglClass = oglLink.getSourceShape()
        destShape: OglClass = oglLink.getDestinationShape()
        self.logger.debug(f'source ID: {srcShape.GetID()} - destination ID: {destShape.GetID()}')

        pyutLink: PyutLink = oglLink.pyutObject

        if pyutLink.linkType == PyutLinkType.INHERITANCE:
            childPyutClass:  PyutClass = cast(PyutClass, srcShape.pyutObject)
            parentPyutClass: PyutClass = cast(PyutClass, destShape.pyutObject)
            childPyutClass.addParent(parentPyutClass)
        else:
            srcPyutClass:  PyutClass = cast(PyutClass, srcShape.pyutObject)
            srcPyutClass.addLink(pyutLink)

    def __furtherCustomizeAssociationLink(self, graphicLink: Element, oglLink: OglAssociation):
        """
        Customize the visual aspects of an Association link

        TODO:  There is no support for this yet.  Need to add it to OglAssociation

        Args:
            graphicLink:  The top level GraphicLink Element
            oglLink:      The current OGL representation of the graphicLink
        """
        center: OglAssociationLabel = oglLink.centerLabel
        src:    OglAssociationLabel = oglLink.sourceCardinality
        dest:   OglAssociationLabel = oglLink.destinationCardinality

        self.__setAssociationLabelPosition(graphicLink, 'LabelCenter', center)
        self.__setAssociationLabelPosition(graphicLink, 'LabelSrc',    src)
        self.__setAssociationLabelPosition(graphicLink, 'LabelDst',    dest)

    def __setAssociationLabelPosition(self, graphicLink: Element, tagName: str, associationLabel: OglAssociationLabel):
        """

        Args:
            graphicLink:  The top level GraphicLink Element
            tagName:      The XML Element name
            associationLabel:  The Ogl association label to update
        """
        labels:  List[Element]   = graphicLink.get_elements(tagName)
        assert len(labels) == 1, 'There can be only one'
        label: Element = labels[0]
        x: int = int(label['x'])
        y: int = int(label['y'])

        self.logger.debug(f'tagName: {tagName} `{associationLabel.text=}`  pos: ({x},{y})')

        associationLabel.oglPosition.x = x
        associationLabel.oglPosition.y = y




from typing import List
from typing import NewType

from dataclasses import dataclass

from untangle import Element

from miniogl.ControlPoint import ControlPoint

from untanglepyut.Types import UntangledOglActors
from untanglepyut.Types import UntangledOglLinks
from untanglepyut.Types import UntangledOglUseCases

UntangledControlPoints = NewType('UntangledControlPoints', List[ControlPoint])

Elements = NewType('Elements', List[Element])

@dataclass
class GraphicInformation:
    """
    Internal Class use to move information from a Graphic XML element
    into Python
    """
    x: int = -1
    y: int = -1
    width:  int = -1
    height: int = -1


"""
Factory methods for our dataclasses
"""


def createUntangledOglLinks() -> UntangledOglLinks:
    return UntangledOglLinks([])


def createUntangledOglUseCases() -> UntangledOglUseCases:
    return UntangledOglUseCases([])


def createUntangledOglActors() -> UntangledOglActors:
    return UntangledOglActors([])


def str2bool(strValue: str) -> bool:
    """
    Converts a known set of strings to a boolean value

    TODO: Put in common place;  Also, in UnTanglePyut

    Args:
        strValue:

    Returns:  the boolean value
    """
    return strValue.lower() in ("yes", "true", "t", "1", 'True')


def secureInteger(x: str):
    if x is not None and x != '':
        return int(x)
    else:
        return 0


def toGraphicInfo(graphicElement: Element) -> GraphicInformation:
    graphicInformation: GraphicInformation = GraphicInformation()

    graphicInformation.x = int(graphicElement['x'])
    graphicInformation.y = int(graphicElement['y'])

    graphicInformation.width  = int(graphicElement['width'])
    graphicInformation.height = int(graphicElement['height'])

    return graphicInformation
